/*********************************************
 *
 * Project : Stalling controller
 * Date    : 24.02.2003
 *
 * Chip type           : AT90S2313
 * Clock frequency     : 1,842300 MHz
 * Memory model        : Tiny
 * Internal SRAM size  : 128
 * External SRAM size  : 0
 * Data Stack size     : 32
*********************************************/

#include <io2313v.h>
#include <macros.h>
#include <stdlib.h>
#include <string.h>

// PI-Reglerkonstanten
#define P_Daempfung_Z 15 						// Dmpfung des P-Reglers Zhler
#define Aus 0
#define Ein 1

// Anzahl Systemticks, bis ein Timeoutfehler generiert wird
#define Motortimeout 1000

// Bits im Fehlerregister
#define Fehler_Nullposition 1		
#define Fehler_Positionierung 2

#define Schrittumrechnung 5 		// Umrechnung (1-9) in absolute Motorposition

#define Kommando_stellen 'W'        // Anstellwinkel ndern
#define Kommando_Fehler  'F'		// Fehlerzustand abfragen
#define Kommando_Null    'N'		// Referenz anfahren
#define Kommando_aus     'A'		// Regelung aus
#define Kommando_ein     'E'		// Regelung ein

// Systemzeit
long SystemTicker = 0;				// Systemuhr

// Motorpositionen
int Motoristposition = 0;			// Aktuelle Position des Motors
int Motorsollposition = 0;			// Sollposition des Motors

// PI-Regler
char Regelgroesse;					// Wert der Ausgangs-PWM
long Stellgroesse;                  // Errechnete Stellgre
char Regelung = Ein;				// Position ausregeln oder nicht ?

// Fehlerergister
char Fehler = 0;                	// Fehlerregister

// Kommandos
char Kommando=0;         			// Empfangenes Kommando
char Wert = 0;                      // Emfangener Parameter

// Rckgabepuffer fr INT -> ASCII-Wandlung
char outpbuf [9];

/*****************************************
 *
 * External Interrupt 0 service routine
 * berwacht den Winkelenkoder des Motors
 *
 *****************************************/
#pragma interrupt_handler int0_isr:2
 void int0_isr (void)
 {
  // Aktuelle Position des Winkelenkoders festhalten
  if (PIND & 8) Motoristposition++; else Motoristposition--;
 }

/*********************************************
 *
 * Timer 0 overflow interrupt service routine
 * MAcht die PI-Regelung. Evtl. Zahlenbereiche
 * korrigieren auf int ?
 *
 * Aufruf ca. jede ms bei 2 Mhz
 * Laufzeit 500us
 *
 *********************************************/
#pragma interrupt_handler timer0_ovf_isr:7
void timer0_ovf_isr (void)
 {
  SystemTicker++;					// Systemuhr erhhen
  
    // P-Anteil bestimmen
    Stellgroesse = ( (Motorsollposition - Motoristposition) * P_Daempfung_Z ); 
  
    // Umrechnen in Ganzzahligen Wert
    if ( abs (Stellgroesse) > 255 ) Regelgroesse = 255; else 
                                    Regelgroesse = abs (Stellgroesse);
  
    // Motor ansteuern, Drehrichtung festlegen               
    if (Stellgroesse < 0) PORTB &= ~16; else 
                          PORTB |= 16;	
   
    // PWM steuern 
	if (Regelung) OCR1L = Regelgroesse; else OCR1L = 0; 

 }

/******************
 *
 * System init
 *
 ******************/
void boot (void)
 {
  CLI ();          // disable all interrupts
 
  // Input/Output Ports initialization
  // Port B initialization
  // Func0=Out Func1=In Func2=In Func3=Out Func4=Out Func5=In Func6=In Func7=In
  // State0=1 State1=T State2=T State3=1 State4=1 State5=T State6=T State7=T
  PORTB=0x19;
  DDRB=0x19;

  // Port D initialization
  // Func0=In Func1=In Func2=In Func3=In Func4=Out Func5=In Func6=In
  // State0=T State1=T State2=T State3=T State4=1 State5=T State6=T
  PORTD=0x10;
  DDRD=0x10;

  // Timer/Counter 0 initialization
  // Clock source: System Clock
  // Clock value: 1842,300 kHz
  TCCR0=0x02;
  TCNT0=0x00;

  // Timer/Counter 1 initialization
  // Clock source: System Clock
  // Clock value: 1842,300 kHz
  // Mode: Ph. correct PWM top=00FFh
  // OC1 output: Inverted
  // Noise Canceler: Off
  // Input Capture on Falling Edge
   TCCR1A=0xC1;
  TCCR1B=0x01;
  TCNT1H=0x00;
  TCNT1L=0x00;
  OCR1H=0x00;
  OCR1L=0x00;					                          // PWM-Startwert, entspricht 0%
  
  // UART initialization
  // Communication Parameters: 8 Data, 1 Stop, No Parity
  // UART Receiver: On
  // UART Transmitter: On
  // UART Baud rate: 9600 / 2 Mhz
  UCR=0x18;
  UBRR=12;

  // External Interrupt(s) initialization
  // INT0: On
  // INT0 Mode: Rising Edge
  // INT1: Off
  MCUCR = 0x03;
  GIMSK = 0x40;
  TIMSK = 0x02;

  SEI ();          	   	   				 	   	  		  // re-enable interrupts

 }

/**************************************************
 *
 * Liefert ein evtl. empfangenes Zeichen, sonst 0
 *
 **************************************************/
char ReceiveByte ()
 {
  if (!(USR & 0x80)) return (0);
  return (UDR);
 }

/**************************************************
 *
 * Sendet ein Byte
 *
 **************************************************/
void SendByte (char zeichen)
 {
  // TXC auf Null setzen durch berschreiben mit 1
  USR |= 64;
    
  // Slaveadresse senden
  UDR = zeichen;
  
  // Warten bis Zeichen gesendet 
  while (!(USR & 64));
 }

/***************************************************************************************************************
 *
 * Integer nach String umwandeln
 *
 ***************************************************************************************************************/
/*
char* inttostr (int twert)
 {
   unsigned short ad,rest;
   char step;
   int wert;
   
   wert = twert;

   memset (outpbuf, 0, 9);	                    // Buffer lschen
   step=0;
   
   if (wert < 0)
    {
	 outpbuf [0]='-';
	 step++;
	 wert *= -1;
	}  
	
   rest = wert;

   if (wert >=10000)	                        // Stellen nacheinander abfragen+zuordnen
   	{
   	 ad=(rest / 10000);
   	 outpbuf [step] = (48+ad);                    // 48 entspricht '0'
   	 rest=rest-(10000*ad);
   	 step++;
   	}
   if (wert >=1000)
   	{
     ad=(rest / 1000);
   	 outpbuf[step]=(ad+48);
   	 rest=rest-(1000*ad);
   	 step++;
   	}
   	if (wert >=100)
   	{
   	 ad=(rest / 100);
   	 outpbuf[step]=ad+48;
   	 rest=rest-(100*ad);
   	 step++;
   	}

   	if (wert >=10)
   	{
   	 ad=(rest / 10);
   	 outpbuf[step]=ad+48;
   	 rest=rest-(10*ad);
   	 step++;
   	}
   	
	outpbuf[step]=(48+rest);
   	outpbuf [step+1]=0;                        // mit Null abschlieen !!
   	step++;
   	return (outpbuf);
 }
*/
 
/**************************************************************************************************************
 *
 * String auf Schnittstelle ausgeben
 *
 **************************************************************************************************************/
void PutString (const char * s)
 {
  while(*s) SendByte (*s++);
 }
 
/**************************************************************************************************************
 *
 * Nummer auf Schnittstelle ausgeben
 *
 **************************************************************************************************************/
/*
void PutNum (void)
 {
  char z = 0;
  
  while (outpbuf [z] != 0)
   {
    SendByte (outpbuf [z]);
    z++;
   }
 }
*/

/***************************************
 *
 * Fhrt den Motor in die Nullposition
 *
 **************************************/
void SucheNull (void)
 {
  // Befindet sich der Motor in der Nullposition ? Dann rausfahren
  if (!(PIND & 64))
   {
     SystemTicker=0;					
     Motorsollposition = 50;
	 while ( (SystemTicker < Motortimeout) && (Motorsollposition != Motoristposition) );
	 if (!(PIND & 64)) Fehler |= Fehler_Nullposition;					  	  // Bit 0 im Fehlerregister setzen
   }
   
  // Nullposition anfahren
  SystemTicker=0;					
  Motorsollposition = -500;
  while ( (Motorsollposition != 0) && (SystemTicker < Motortimeout) )
   // Ist die Ruheposition erreicht ? Dann Motor abschalten durch Setzen der Positionen
   if (!(PIND & 64))
    {
     Motorsollposition = 0;
     Motoristposition = 0;
	 OCR1L = 0;      		        					// Motor aus
    }
  
  // Timeout aufgetreten ?
  if (SystemTicker >= Motortimeout)
   {               
    Motorsollposition=0;
    Motoristposition=0;
    OCR1L = 0;      		        					// Motor aus
    Fehler |= Fehler_Nullposition;					  	  // Bit 0 im Fehlerregister setzen
   } 

 // Position 100 anfahren. Dort greift die Druckplatte
  SystemTicker=0;					
  Motorsollposition = 100;
  while ( (Motoristposition != 100) && (SystemTicker < Motortimeout) );

  Motorsollposition = 0;
  Motoristposition = 0;
  OCR1L = 0;      		        					// Motor aus

  // Timeout aufgetreten ?
  if (SystemTicker >= Motortimeout) Fehler |= Fehler_Nullposition;					  	  // Bit 0 im Fehlerregister setzen
 } 
 
/******************
 *
 * Main
 *
 ******************/
void main (void)
 {
  
  boot (); 			 		// System init
  
  PutString ("Start");	    // Startmeldung ausgeben
 
  // Hauptschleife
  while (1)
   { 
     // Entgegennehmen von Kommandos
     Kommando = ReceiveByte ();
     if (Kommando != 0)
      {          
       // Gltiges Kommando wurde empfangen
       if (Kommando == Kommando_stellen)
        {
         outpbuf [0] = 0;
         while (outpbuf [0] == 0) outpbuf [0] = ReceiveByte (); 
         outpbuf [1] = 0;
         while (outpbuf [1] == 0) outpbuf [1] = ReceiveByte (); 
		 
		 // Der Wert (1-9) enthlt jetzt den Anstellwinkel und mu umgerchnet werden in Motorschritte
         Motorsollposition =  (long)(((outpbuf [0]-48)*10) + (outpbuf [1]-48)) * Schrittumrechnung;
		 
		  // Ergebniss anzeigen
/*
		  inttostr (Motorsollposition);
		  PutNum ();
		  PutString (":");
		  inttostr (Motoristposition);
		  PutNum ();
		  PutString (" # ");
*/		 
		 // Warten, bis diese Position erreicht wurde, oder ein Timeout auftritt
		 SystemTicker=0;					
  		 while ( (Motorsollposition != Motoristposition) && (SystemTicker < Motortimeout) );
  
  	     // Timeout aufgetreten ?
  		 if (SystemTicker >= Motortimeout)
   		  {               
    	   Motorsollposition=0;
    	   Motoristposition=0;
    	   Fehler |= Fehler_Positionierung;					  	  // Bit 1 im Fehlerregister setzen
   		  }

		  // Ergebniss anzeigen
		  SendByte (Fehler+48);
/*
		  inttostr (Motorsollposition);
		  PutNum ();
		  PutString (":");
		  inttostr (Motoristposition);
		  PutNum ();
		  outpbuf [0]=10;
		  outpbuf [1]=13;
		  outpbuf [2]=0;
		  PutNum ();
*/		  
        } // Von if (Kommando == Kommando_stellen)
     
       // Gltiges Kommando wurde empfangen
       if (Kommando == Kommando_Fehler) SendByte (Fehler+48);
 
       // Gltiges Kommando wurde empfangen
       if (Kommando ==  Kommando_Null) 
	    {
		 Fehler = 0;
         SucheNull (); 
		 SendByte (Fehler+48);
		} 

       // Gltiges Kommando wurde empfangen
       if (Kommando ==  Kommando_aus) 
	    {
		 PORTB &= ~1;
		 Regelung = Aus;
		 SendByte ('A');
		} 

		       // Gltiges Kommando wurde empfangen
       if (Kommando ==  Kommando_ein) 
	    {
		 Regelung = Ein;
		 PORTB |= 1;
		 SendByte ('E');
		} 

      } // Von  if (Kommando != 0)
        
   }

 }
 
 